//-------------------------------------------------------------------------
// play-v6 - music player for ATmega328P @ 16 MHz
// user interface (except serial)
// by Enrico Colombini <erix@erix.it>
// license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
// http://creativecommons.org/licenses/by-sa/4.0/
//-------------------------------------------------------------------------

#include "interface.h"

//-------------------------------------------------------------------------
// local types

//state of all buttons (note: bit=1 when button is pressed)
typedef struct {
    uint8_t state;          //accepted state (bits) for all buttons
    uint8_t lastRead;       //last read state (bits)
    uint8_t debounce;       //debounce counter
} ButtonsInfo;

//-------------------------------------------------------------------------
// local variables

static ButtonsInfo buttonsInfo;

//-------------------------------------------------------------------------
// init port bits and state for all buttons

void InitInterface(void)
{
    //set button port bits as input
    BUTTON_PORT_DDR &= (~ ALL_BUTTONS_MASK);
    //enable pull-up resistors
    BUTTON_PORT |= ALL_BUTTONS_MASK;

    //read current hardware state
    //NOTE: this can be wrong for inputs with a weak pull-up or a capacitor,
    //call CopyHardwareButtonState later to get the actual input state
    buttonsInfo.state = (~BUTTON_PORT_PIN) & ALL_BUTTONS_MASK;
    //but consider it unstable, force debounce
    buttonsInfo.lastRead = buttonsInfo.state ^ ALL_BUTTONS_MASK;
    buttonsInfo.debounce = 0;
}

//-------------------------------------------------------------------------
// update state of all buttons, handling debounce:
// all buttons must have been stable for BUTTON_DEBOUNCE_COUNT calls
// (call repeatedly with appropriate timing, e.g. 256 calls in ~8 ms)
// return a byte with a bit set for each button that has changed state

uint8_t UpdateButtons(void)
{
    //set bits here for each button that changed state, after debounce
    uint8_t changed = 0x00;

    //read current input bits (1 = button pressed, for convenience)
    uint8_t inputs = (~ BUTTON_PORT_PIN) & ALL_BUTTONS_MASK;

    if (inputs != buttonsInfo.lastRead)
    {
        //unstable, reset debounce counter
        buttonsInfo.debounce = 0;
    }
    else
    {
        //inputs are the same as last reading, debounce
        //(counts even when the button is stable, no harm done)
        buttonsInfo.debounce++;
        if (buttonsInfo.debounce >= BUTTON_DEBOUNCE_COUNT)
        {
            //debounce complete, get changed bits
            changed = inputs ^ buttonsInfo.state;
            //store new state (state, inputs, lastRead will be equal)
            buttonsInfo.state = inputs;
            //reset debounce counter
            buttonsInfo.debounce = 0;
        }
    }

    //remember current input bits in any case
    buttonsInfo.lastRead = inputs;

    return changed;
}

//-------------------------------------------------------------------------
// return true if the given button is currently pressed (debounced)

bool IsButtonPressed(uint8_t buttonMask)
{
    return ((buttonsInfo.state & buttonMask) != 0x00);
}

//-------------------------------------------------------------------------
// return true if the given button has just been pressed,
// must be called with changed = result from last call to UpdateButtons

bool HasButtonBeenPressed (uint8_t changed, uint8_t buttonMask)
{
    return ((changed & buttonsInfo.state & buttonMask) != 0x00);
}

//-------------------------------------------------------------------------
// set stored state for buttons
// (e.g. set to pressed to ensure release debounce before pressing)

void SetButtonState(uint8_t state)
{
    buttonsInfo.state = state & ALL_BUTTONS_MASK;
    buttonsInfo.lastRead = buttonsInfo.state;
    buttonsInfo.debounce = 0;
}

//-------------------------------------------------------------------------
// copy hardware input state to current state

void CopyButtonHardwareState(void)
{
    buttonsInfo.state = (~BUTTON_PORT_PIN) & ALL_BUTTONS_MASK;
}

//-------------------------------------------------------------------------
