//-------------------------------------------------------------------------
// play-v6 - music player for ATmega328P @ 16 MHz
// player and music data interpreter
// by Enrico Colombini <erix@erix.it>
// license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
// http://creativecommons.org/licenses/by-sa/4.0/
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// NOTE:
// this file compiles to a single function with no calls,
// must be compiled with -Os to work correctly
//-------------------------------------------------------------------------

#include "play.h"
#include "play_v6.h"
#include "pwm.h"
#include "interface.h"
#include "music.h"
#include "instrum.h"
#include "tables.h"
#include "userial.h"
#include <avr/io.h>
#include <avr/pgmspace.h>

//-------------------------------------------------------------------------
// local constants

#define NUM_VOICES              6           //hard-coded, do not change

#define DEFAULT_TEMPO           60u         //qpm (quarter notes per minute)

#define MAX_VOLUME              255u        //max value for voice volume

//task selector mask, hand-picked to match every 128 us (approx.)
//corresponding to 4 play loop cycles at 16 MHz, 512 CPU cycles/loop,
//ADSR update rate for each voice is 128 us * 6 = 768 us
#define TASK_SELECTOR_MASK      (4 - 1)     //0x00000011

//task selection values (after masking)
enum TaskSelector{ TASK_ADSR = 0x00, TASK_SERVE_REQUESTS = 0x01 };

//state of ADSR state machine
enum AdsrState
    { ADSR_START, ADSR_ATTACK, ADSR_DECAY, ADSR_SUSTAIN, ADSR_RELEASE, ADSR_END };

//-------------------------------------------------------------------------
// local types

//24-bit position in a wave table in PROGMEM,
//for convenience it can be accessed in three different ways
//NOTE: the table must be 256-byte aligned and no larger than 256 bytes
typedef union {
    //wave position as 16.8 value
    struct {
        uint8_t     frac;           //fractional part
        const int8_t *ptr;          //pointer to current sample
    };
    //same wave position as 8.16
    struct {
        uint16_t    pos;            //position in table (8.8)
        uint8_t     high;           //(never incremented during update)
    };
    //same wave position as 8.8.8
    struct {
        uint8_t     frac_alias;
        uint8_t     idx;            //index in table
        uint8_t     high_alias;
    };
} WavePos;

//current ADSR position (8.8 index in table)
typedef union {
    //value as 16 bits
    uint16_t        value16;
    //same value as 8.8
    struct {
        uint8_t     frac;           //fractional part
        uint8_t     idx;            //index in table
    };
} AdsrPos;

//status of a voice (time-critical part, within reach of Y frame pointer)
typedef struct
{
    WavePos         wavePos;        //current position (16.8) in wave samples
    uint16_t        waveStep;       //wavePos.pos step (8.8), determines pitch
    uint8_t         volume;         //current playing volume for samples
    uint8_t         reqVolume;      //volume to be set at wave-wrap (set by ADSR)
    uint8_t         noteIdx;        //index of currently playing note (for sync)
} Voice;

//status of the ADSR for a voice (less time-critical part of voice)
typedef struct {
    enum AdsrState  state;          //current state
    uint8_t         reqNoteIdx;     //index of requested note
    uint8_t         noteVolume;     //music volume of requested/current note
    uint8_t         currentLevel;   //current volume level before noteVolume
    uint8_t         refLevel;       //reference volume level, used in release
    AdsrPos         pos;            //current position in table for current state
} VoiceAdsr;

//-------------------------------------------------------------------------
// macro: update wave position for a voice, read sample, apply volume,
// write processed sample into dest as unsigned value
// vp is a pointer to a Voice structure
// NOTE: highly optimized code, always look at generated asm when touching it
// NOTE: implemented as a macro (instead of inline) to be suitable for -Os
//
// - cycle count min: 35 (timer 0), 36 (timer 1,2)
// - cycle count max: 41 (timer 0), 41 (timer 1,2)

#define UpdateVoiceSample(vp, dest) \
do { \
    /*advance to next sample (8. idx + .8 frac):*/ \
    (vp)->wavePos.pos += (vp)->waveStep; \
    /*wrap at end of wave*/ \
    /*NOTE: this works because (WAVE_CYCLE + largest note step.hi) < 256*/ \
    if ((vp)->wavePos.idx >= WAVE_CYCLE) \
    { \
        (vp)->wavePos.idx -= WAVE_CYCLE; \
        /*change volume here (upward zero-crossing) for minimum disruption*/ \
        (vp)->volume = (vp)->reqVolume; \
    } \
    /*get current sample*/ \
    int8_t sample = pgm_read_byte_near((vp)->wavePos.ptr); \
    /*apply volume (multiply signed sample by unsigned volume, keep the high byte)*/ \
    /*note: implicit promotion to 16 bits, the compiler rightly produces MULSU*/ \
    sample = ((sample * (vp)->volume) >> 8) & 0xff; \
    /*convert to unsigned for timer comparator, assign to destination*/ \
    (dest) = ((uint8_t)sample) + ZERO_LEVEL; \
} while(0)

//-------------------------------------------------------------------------
// play from music data until stop requested
// this is a long all-in-one function with tight timing requirements;
// a play cycle is made of two PWM cycles:
// - PWM cycle 1: compute and set samples for all voices
// - PWM cycle 2: read and execute music commands, update ADSRs, aux. tasks
// (in this version, each PWM plays a single voice)
// return true if END of music reached, false if STOP command received from serial

enum PlayState Play(const uint8_t *music)
{
    //---------------------------------------------------------------------
    // play loop local variables and initialization
    // (local vars are much faster because of Y frame pointer)
    // voices are initialized to silence, waveStep will be set by 'note' command

    //local variables are inside struct v to enforce their order in stack frame:
    //time-critical data must be first (to be within 63 bytes for LDD,Y)
    struct {
        //playing status of the voices (time-critical, must be first)
        //NOTE: must stay within 63 bytes of frame pointer (Y register)
        Voice voices[NUM_VOICES];

        //number of PWM cycles to make 1/64 measure with current tempo
        uint16_t tempo64Step;

        //countdown from tempoStep64 to 0
        uint16_t tempo64Count;

        //number of 1/64 to wait before reading next data
        uint8_t musicWait64;

        //free-running selector for ADSR and other tasks
        uint8_t taskSelector;

        //current ADSR voice (0..NUM_VOICES)
        uint8_t adsrVoiceIdx;

        //pointer to music data (avoids 'music' being placed on top of frame)
        const uint8_t *musicPtr;

        //play loop state
        enum PlayState state;

        //ADSR status of the voices, separated to keep size of voices[] <64
        //(least time-critical, can be last)
        VoiceAdsr adsrs[NUM_VOICES];

        //play/stop button debounce counter
        uint8_t buttonDebounce;

    } v; //(struct enforcing order of local vars in stack frame)

    //init voice data
    uint8_t voiceIdx;
    for (voiceIdx = 0; voiceIdx < NUM_VOICES; ++voiceIdx)
    {
        //time-critical data
        Voice *vp = &v.voices[voiceIdx];
        //default = first instrument; get high byte of its address
        vp->wavePos.high = pgm_read_byte_near((uint16_t)(&instruments[0]) + 1);
        //set voice pos to minimize time to wave-wrap, for faster first attack
        //(remove -1 to test for worst case: all voices will wrap at first cycle)
        vp->wavePos.idx = WAVE_CYCLE - 1;
        vp->wavePos.frac = 0;
        vp->waveStep = 0;
        vp->volume = 0;
        vp->reqVolume = 0;
        //invalid index to be sure it will not be used as sync reference by ADSR
        vp->noteIdx = 0xff;

        //ADSR data
        VoiceAdsr *ap = &v.adsrs[voiceIdx];
        ap->state = ADSR_END;
        ap->reqNoteIdx = 0;
        ap->noteVolume = 0;
        ap->currentLevel = 0;
        ap->refLevel = 0;
        ap->pos.frac = 0;
        ap->pos.idx = 0;
    }

    //(tempo64Steps[] is 1-based)
    v.tempo64Step = pgm_read_word_near(&tempo64Steps[DEFAULT_TEMPO - 1]);
    //start countdown to 1/64
    v.tempo64Count = v.tempo64Step;
    //not currently waiting, will read music commands
    v.musicWait64 = 0;
    //task selector is free-running
    v.taskSelector = 0;
    //ready on ADSR for first  voice
    v.adsrVoiceIdx = 0;
    //ready on start of music
    v.musicPtr = music;

    //---------------------------------------------------------------------
    // prepare to start play loop

    //uncomment this to send frame pointer address via serial,
    //to measure RAM space available for music buffer
    /*
    uint16_t lastAddress = (uint16_t)&v;
    UART0_SendByte((lastAddress >> 8) & 0xff);
    UART0_SendByte(lastAddress & 0xff);
    */

    v.state = PLAY_PLAYING;
    v.buttonDebounce = 0;

    //be sure to wait for sync to have a full PWM cycle available
    //even if the optimizer moves instructions here
    ResetPWMSync();
    SyncToPWM();

    //NOTE: the optimizer can move some instructions here;
    //always count cycles to be sure the first sync will not be missed

    //---------------------------------------------------------------------
    // time-critical play loop starts here

    while (v.state < PLAY_EXIT_DONE)
    {
        //-----------------------------------------------------------------
        // play loop phase 1: compute and play wave samples

        //start of first PWM cycle: play samples
        SyncToPWM();

        //update and play samples for all voices
        //(time-critical, uses optimized macro)
        //NOTE: timers are staggered by PWM_STAGGER_CYCLES (defined in pwm.h)

        //voices set on PWM timer 0
        UpdateVoiceSample(&v.voices[0], PWM1Set);
        UpdateVoiceSample(&v.voices[1], PWM2Set);

        //voices set on PWM timer 1
        UpdateVoiceSample(&v.voices[2], PWM3Set);
        UpdateVoiceSample(&v.voices[3], PWM4Set);

        //voices set on PWM timer 2
        UpdateVoiceSample(&v.voices[4], PWM5Set);
        UpdateVoiceSample(&v.voices[5], PWM6Set);

        //NOTE: voice 6 can finish after timer 0 cycle has expired,
        //but PWM timers are staggered, so timer 2 is still counting

        //DebugSendMaxTimer0(); //(for testing, only reliable with 1..5 voices)

        //-----------------------------------------------------------------
        // first PWM cycle (voices) ends here; second cycle starts
        // (needed but could be executed a bit after first PWM cycle ended)
        SyncToPWM();

        //-----------------------------------------------------------------
        // play loop phase 2: update 1/64 'clock' and wait counter

        //check if 1/64 of measure expired
        if (--v.tempo64Count == 0)
        {
            //reload 1/64 'clock' countdown
            v.tempo64Count = v.tempo64Step;

            //update remaining music waiting time (in 1/64), if any
            //if zero, it will be reloaded by the WAIT command in music
            if (v.musicWait64 > 0)
            {
                v.musicWait64--;
            }
        }

        //-----------------------------------------------------------------
        // play loop phase 3a: if not waiting, get and execute a music command
        // (one command per cycle, until a WAIT command is encountered)

        if ((v.musicWait64 == 0) && (v.state == PLAY_PLAYING))
        {
            //read a command from music, advance
            const uint8_t cmdByte = *v.musicPtr++;

            //look at main music command bits
            switch (cmdByte >> MUSCMD_MAIN_SHIFT)
            {
                case MUSCMD_WAIT:
                    //bits 5..0 = number of 1/64 to wait (0 means 64)
                    v.musicWait64 = cmdByte; //(no need to & 0x3f, upper bits are 0)
                    if (v.musicWait64 == 0)
                    {
                        v.musicWait64 = 64;
                    }
                    break;

                case MUSCMD_NOTE:
                    {
                        //read argument
                        const uint8_t argByte = *v.musicPtr++;

                        //bits 2..0 = voice (check for safety)
                        uint8_t voiceIdx = argByte & 0x07;
                        if (voiceIdx >= NUM_VOICES)
                        {
                            voiceIdx = 0;
                        }
                        VoiceAdsr *ap = &v.adsrs[voiceIdx];
                        //bits 5..0 of first byte = note
                        ap->reqNoteIdx = cmdByte & 0x3f;
                        //bits 7..3 of arg = volume (0..31 -> 0..248, lower bits set to 0)
                        //(could use table to expand to full volume range)
                        ap->noteVolume = argByte & 0xf8;
                        //tell current note to stop
                        v.voices[voiceIdx].reqVolume = 0;
                        //tell ADSR to start the new note when the current one will be silent
                        ap->state = ADSR_START;
                    }
                    break;

                case MUSCMD_RELEASE:
                    {
                        //bits 2..0 = voice (check for safety)
                        uint8_t voiceIdx = cmdByte & 0x07;
                        if (voiceIdx >= NUM_VOICES)
                        {
                            voiceIdx = 0;
                        }
                        VoiceAdsr *ap = &v.adsrs[voiceIdx];
                        //tell ADSR to start release from current level
                        ap->refLevel = ap->currentLevel;
                        ap->pos.value16 = 0;
                        ap->state = ADSR_RELEASE;
                    }
                    break;

                case MUSCMD_OTHER:
                    //other music commands, look at full music command
                    switch (cmdByte)
                    {
                        case MUSCMD_END:
                            //exit player when current notes finish playing:
                            //tell ADSRs to skip sustain,
                            //wait for all voices to become silent, then exit
                            //(will also stop reading music)
                            v.state = PLAY_END_REQUEST;
                            break;

                        case MUSCMD_TEMPO:
                            {
                                //read argument (quarter notes per minute)
                                const uint8_t argByte = *v.musicPtr++;
                                //note: tempo64Steps[] is 1-based
                                v.tempo64Step = pgm_read_word_near(&tempo64Steps[argByte - 1]);
                                //reload 1/64 'clock' countdown with new value (it had just been loaded)
                                v.tempo64Count = v.tempo64Step;
                            }
                            break;

                        case MUSCMD_INSTRUMENT:
                            {
                                //read argument
                                const uint8_t argByte = *v.musicPtr++;
                                //bits 2..0 = voice (check for safety)
                                uint8_t voiceIdx = argByte & 0x07;
                                if (voiceIdx >= NUM_VOICES)
                                {
                                    voiceIdx = 0;
                                }
                                //bits 7..4 = instrument
                                uint8_t instrumIdx = (argByte >> 4) & 0x0f;

                                //put high byte of instrument address into voice
                                v.voices[voiceIdx].wavePos.high =
                                    pgm_read_byte_near((uint16_t)(&instruments[instrumIdx]) + 1);
                            }
                            break;

                        default:
                            //unrecognized commands act as MUSCMD_END
                            //to help prevent music buffer overflow
                            v.state = PLAY_END_REQUEST;
                            break;

                    } //(MUSCMD_OTHER)
                    break;

                default:
                    //(unrecognized commands already handled in MUSCMD_OTHER sub-switch)
                    break;

            } //(main command switch)

        } //(if (musicWait64 == 0) and playing, i.e. executing commands)

        // NOTE: continues with 'else' in phase 3b

        //-----------------------------------------------------------------
        // play loop phase 3b: if waiting,
        // (i.e. no music commands to execute, or music end has been reached)
        // handle ADSR and housekeeping functions

        else //(musicWait64 != 0) || (v.state != PLAY_PLAYING)
        {
            //increment task selector blindly, let it wrap around
            //this is not accurate, cycles are skipped while executing commands
            //(especially for very fast tempo) but should be good enough
            v.taskSelector++;
            uint8_t currentTask = v.taskSelector & TASK_SELECTOR_MASK;

            if (currentTask == TASK_ADSR)
            {
                //ADSR time tick: serve one voice per tick
                Voice *vp = &v.voices[v.adsrVoiceIdx];
                VoiceAdsr *ap = &v.adsrs[v.adsrVoiceIdx];

                switch (ap->state)
                {
                    case ADSR_START:
                        //do nothing if previous note is still playing (wait), else:
                        if (vp->volume == 0)
                        {
                            //voice is silent:
                            //sync wave pos if same note to avoid phase problems
                            //(syncing to the first voice found is enough,
                            //also no need to sync high byte, it does not change)
                            if (ap->reqNoteIdx == v.voices[0].noteIdx) {
                                vp->wavePos.pos = v.voices[0].wavePos.pos;
                            }
                            else if (ap->reqNoteIdx == v.voices[1].noteIdx) {
                                vp->wavePos.pos = v.voices[1].wavePos.pos;
                            }
                            else if (ap->reqNoteIdx == v.voices[2].noteIdx) {
                                vp->wavePos.pos = v.voices[2].wavePos.pos;
                            }
                            else if (ap->reqNoteIdx == v.voices[3].noteIdx) {
                                vp->wavePos.pos = v.voices[3].wavePos.pos;
                            }
                            else if (ap->reqNoteIdx == v.voices[4].noteIdx) {
                                vp->wavePos.pos = v.voices[4].wavePos.pos;
                            }
                            else if (ap->reqNoteIdx == v.voices[5].noteIdx) {
                                vp->wavePos.pos = v.voices[5].wavePos.pos;
                            }

                            //start new note
                            vp->waveStep = pgm_read_word_near(&notes[ap->reqNoteIdx]);
                            //code of actually playing note is needed for phase sync
                            vp->noteIdx = ap->reqNoteIdx;
                            //start of attack (currently from 0, ignoring current level)
                            ap->pos.value16 = 0;
                            ap->state = ADSR_ATTACK;
                        }
                        break;

                    case ADSR_ATTACK:
                        //linear attack (uses an imaginary linear table, volume = idx)
                        //(an ADSR cycle has already passed, so increment first)
                        {
                            //get attack step for current instrument
                            InstrumentConstPtr ip = (InstrumentConstPtr)(vp->wavePos.high << 8);
                            const uint16_t attackStep = pgm_read_word_near(&ip->attackStep);

                            //advance 8.8 position
                            uint8_t preIdx = ap->pos.idx;
                            ap->pos.value16 += attackStep;
                            if (ap->pos.idx < preIdx)
                            {
                                //wrapped: end of attack, set level to max
                                ap->currentLevel = 255;
                                //reset decay table position
                                ap->pos.value16 = 0;
                                ap->state = ADSR_DECAY;
                            }
                            else
                            {
                                //set level to current attack value
                                ap->currentLevel = (uint16_t)ap->pos.idx;
                            }
                            //set wave volume
                            vp->reqVolume = (ap->noteVolume * ap->currentLevel) >> 8;
                        }
                        break;

                    case ADSR_DECAY:
                        //exponential decay to sustain volume
                        {
                            //get decay step, sustain level for current instrument
                            InstrumentConstPtr ip = (InstrumentConstPtr)(vp->wavePos.high << 8);
                            const uint16_t decayStep = pgm_read_word_near(&ip->decayStep);
                            const uint8_t sustainLevel = pgm_read_byte_near(&ip->sustainLevel);

                            //advance 8.8 position
                            uint8_t preIdx = ap->pos.idx;
                            ap->pos.value16 += decayStep;
                            if (ap->pos.idx < preIdx)
                            {
                                //wrapped: end of decay, set level to sustain
                                ap->currentLevel = sustainLevel;
                                ap->state = ADSR_SUSTAIN;
                            }
                            else
                            {
                                //get current decay value, scale to max-sustain range
                                uint8_t dv = pgm_read_byte_near(adsrDecayTable + ap->pos.idx);
                                dv = (dv * (255 - sustainLevel)) >> 8;
                                ap->currentLevel = dv + sustainLevel;
                            }
                            //set wave volume
                            vp->reqVolume = (ap->noteVolume * ap->currentLevel) >> 8;
                        }
                        break;

                    case ADSR_SUSTAIN:
                        if (v.state == PLAY_END_REQUEST)
                        {
                            //end of music: start release from current level
                            ap->refLevel = ap->currentLevel;
                            ap->pos.value16 = 0;
                            ap->state = ADSR_RELEASE;
                        }
                        //else just stay here until RELEASE command
                        break;

                    case ADSR_RELEASE:
                        //exponential decay from reference (current) volume to zero
                        //(not entirely correct, release time should depend on volume)
                        {
                            //get release step for current instrument
                            InstrumentConstPtr ip = (InstrumentConstPtr)(vp->wavePos.high << 8);
                            const uint16_t releaseStep = pgm_read_word_near(&ip->releaseStep);

                            //advance 8.8 position
                            uint8_t preIdx = ap->pos.idx;
                            ap->pos.value16 += releaseStep;
                            if (ap->pos.idx < preIdx)
                            {
                                //wrapped: end of release, set wave volume to zero
                                vp->reqVolume = 0;
                                ap->state = ADSR_END;
                            }
                            else
                            {
                                //get current release value, scale to ref-0 range
                                uint8_t dv = pgm_read_byte_near(adsrDecayTable + ap->pos.idx);
                                ap->currentLevel = (dv * ap->refLevel) >> 8;
                                //set wave volume
                                vp->reqVolume = (ap->noteVolume * ap->currentLevel) >> 8;
                            }
                        }
                        break;

                    default:
                        //ADSR_END falls here
                        break;

                } //(switch(ap->state)

                //next voice for next tick
                v.adsrVoiceIdx++;
                if (v.adsrVoiceIdx >= NUM_VOICES)
                {
                    v.adsrVoiceIdx = 0;
                }

            } //(currentTask == TASK_ADSR)

            else if (currentTask == TASK_SERVE_REQUESTS)
            {
                bool exitNow = false;

                //check serial for immediate exit request
                if (UART0_IsRxDataAvailable())
                {
                    //a char has been received from the serial port
                    uint8_t cmd = UART0_ReceiveByteNoCheck();

                    if (cmd == SERIAL_CMD_QUERY)
                    {
                        //status query, send reply
                        //(no check for free transmitter,
                        //could be a problem if queried too fast)
                        UART0_SendByteNoCheck(SERIAL_REPLY_PLAYING);
                    }
                    else if (cmd == SERIAL_CMD_PLAY_TOGGLE)
                    {
                        v.state = PLAY_EXIT_STOP;
                        exitNow = true;
                    }
                    else if (cmd == SERIAL_CMD_NEXT)
                    {
                        v.state = PLAY_EXIT_NEXT;
                        exitNow = true;
                    }
                    else if (cmd == SERIAL_CMD_PREV)
                    {
                        v.state = PLAY_EXIT_PREV;
                        exitNow = true;
                    }

                    //(ignore other serial commands)

                    if (exitNow)
                    {
                        //acknowledge accepted serial exit command
                        UART0_SendByteNoCheck(SERIAL_REPLY_OK);
                    }
                }

                //no serial data received,
                //serve end request only when all voices are silent
                else if ((v.state == PLAY_END_REQUEST)
                         && (v.voices[0].volume == 0)
                         && (v.voices[1].volume == 0)
                         && (v.voices[2].volume == 0)
                         && (v.voices[3].volume == 0)
                         && (v.voices[4].volume == 0)
                         && (v.voices[5].volume == 0))
                {
                    //tell play loop to terminate
                    v.state = PLAY_EXIT_DONE;
                }

                //no other requests, check Play/Stop button
                else
                {
                    //simplified debounce: look at all buttons state
                    uint8_t pressedButtons = (~ BUTTON_PORT_PIN) & ALL_BUTTONS_MASK;

                    if (pressedButtons != 0x00)
                    {
                        //at least one button is pressed
                        if (v.buttonDebounce >= BUTTON_DEBOUNCE_COUNT)
                        {
                            //released and debounced: valid exit request
                            if ((pressedButtons & NEXT_BUTTON_MASK) != 0)
                            {
                                v.state = PLAY_EXIT_NEXT;
                            }
                            else if ((pressedButtons & PREV_BUTTON_MASK) != 0)
                            {
                                v.state = PLAY_EXIT_PREV;
                            }
                            else
                            {
                                v.state = PLAY_EXIT_STOP;
                            }
                            //in any case, exit immediately
                            exitNow = true;
                        }
                        else
                        {
                            //not fully debounced: reset counter
                            v.buttonDebounce = 0;
                        }
                    }
                    else
                    {
                        //no button pressed: advance release debounce counter
                        if (v.buttonDebounce < BUTTON_DEBOUNCE_COUNT)
                        {
                            v.buttonDebounce++;
                        }
                    }
                } //(Play/Stop button)

                //serial or button may have issued an immediate exit request
                if (exitNow)
                {
                    //stop playing immediately, truncate all notes
                    PWM1Set = ZERO_LEVEL;
                    PWM2Set = ZERO_LEVEL;
                    PWM3Set = ZERO_LEVEL;
                    PWM4Set = ZERO_LEVEL;
                    PWM5Set = ZERO_LEVEL;
                    PWM6Set = ZERO_LEVEL;
                    //v.state already set to a value >= PLAY_EXIT_DONE
                    //(play loop will terminate)
                }

            } //(currentTask == TASK_SERVE_REQUESTS)

            // [ could serve other tasks here (LED? MIDI?) ]

        } //(end of phase 3a (musicWait64 == 0) or 3b (musicWait64 != 0))

        //-----------------------------------------------------------------
        // all work done for this play loop iteration

        //DebugSendMaxTimer0(); //(test measurement)

    } //(time-critical play loop)

    //return PLAY_EXIT_DONE, PLAY_EXIT_STOP, PLAY_EXIT_NEXT or PLAY_EXIT_PREV
    return (v.state);
}

//-------------------------------------------------------------------------
