---------------------------------------------------------------------------
-- instrument generator for play-v6
-- created Mar 2014 by Enrico Colombini <erix@erix.it>
-- license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
-- http://creativecommons.org/licenses/by-sa/4.0/
---------------------------------------------------------------------------

-- requirements:
--   SoX (http://sox.sourceforge.net) must be on the executable path
--   io.popen() is required (usually available in major operating systems)

---------------------------------------------------------------------------

version = '- instruments 1.0 - (C) 2014 Enrico Colombini'
usage = '- usage: instrument c      -- generate tables as C source code\n' ..
        '         instrument list   -- generate instrument list for musicomp\n' ..
        '         instrument <name> -- generate audible sample for <name>\n'

---------------------------------------------------------------------------
-- constants

-- number of samples in wave table, 
-- very near to G3 (195.9977 Hz) sampled at 44100 Hz
-- convenient for testing (does not matter for play pitch precision)
waveLength = 225

-- number of instruments in table
totalInstruments = 16

-- highest level in ADSR tables
adsrMaxLevel = 255
-- number of entries in ADSR tables
adsrTableSize = 256
-- time between ADSR steps in player (approx.)
adsrTimeStep = 768e-6

-- normalizing level to minimize clipping when converting to integers
normalizeLevel = 0.998

-- frequency used to generate test samples (must be supported by .wav files)
sampleFrequency = 44100
-- base frequency, for testing
baseNoteFreq = sampleFrequency / waveLength

-- duration (s) of silence between sounds in test .wav
testSilenceDuration = 0.5
-- limit for all durations (s), will be clamped
maxDuration = 10.0

-- frequencies of test notes produced in .wav file (same note in 5 octaves)
testFrequencies = { 
    baseNoteFreq / 4, baseNoteFreq / 2, 
    baseNoteFreq, 
    baseNoteFreq * 2, baseNoteFreq *4
}

---------------------------------------------------------------------------
-- instrument data (this could be moved to an external file)
-- this is an ordered list,
-- each instrument is defined by the relative level (in dB) 
-- of the fundamental frequency [1] and of its harmonics [2..]
-- and its ADSR parameters:
--
-- name = {
--     [] dB level of each harmonic, from fundamental wave up
--     attackDuration = duration of linear attack (s),
--     decayDuration = duration of exponential decay (s),
--     sustainLevel = constant level after decay (0.0..1.0),
--     releaseDuration = duration of exponential release (s),
--     testReleaseTime = time of simulated release in test sound (s),
-- }
--
-- NOTE: remember to update m_compiler when instrument positions change

instruments = {
    {
        name = 'sine',
        0.0,
        attackDuration = 0.0,
        decayDuration = 0.0,
        sustainLevel = 1.0,
        releaseDuration = 0.0,
        testReleaseTime = 1.0,
    },
    
    {
        name = 'guitar',
        -12.0, -8.3, -33.1, -25.0, -34.0, -31.6, -31.6, -36.4, -45.2, -42.6,
        -31.1, -41.4, -35.4, -47.5, -42.7, -34.2, -51.9, -54.0, -46.4, -48.3,
        attackDuration = 0.020,
        decayDuration =  2.0,
        sustainLevel = 0.01,
        releaseDuration = 0.040,
        testReleaseTime = 2.0,
    },
    
    {
        name = 'guitar_low',
        -13.1, -27.8, -28.1, -37.4, -31.7, -33.4, -57.7, -48.1, -44.5, -46.3,
        -47.7, -47.2, -44.6,
        attackDuration = 0.020,
        decayDuration =  2.0,
        sustainLevel = 0.01,
        releaseDuration = 0.040,
        testReleaseTime = 2.0,
    },

    {
        name = 'piano',
        -9.1, -45.1, -32.3, -46.6, -53.3, -49.0, -64.8, -70.4,
        attackDuration = 0.025,
        decayDuration = 2.0,
        sustainLevel = 0.080,
        releaseDuration = 0.120,
        testReleaseTime = 1.5,          
    },
    
    {
        name = 'organ',
        -14.9, -19.9, -20.9, -27.0, -50.3, -59.9, -72.5, -69.8, -77.1,
        attackDuration = 0.040,
        decayDuration = 0.080,
        sustainLevel = 0.9,
        releaseDuration = 0.090,
        testReleaseTime = 0.8,          
    },
    
    {
        name = 'flute',
        -0.0, -30.0, -30.0, -3.0, -18.0,
        attackDuration = 0.080,
        decayDuration = 0.2,
        sustainLevel = 0.7,
        releaseDuration = 0.3,
        testReleaseTime = 0.7,
    },
    
    {
        name = 'violin',
        -13.4, -20.8, -14.2, -11.2, -22.3, -18.8, -24.0, -36.3, -32.4, -43.8,
        -37.6, -43.1, -49.1, -50.0, -49.7, -49.3, -62.4, -65.4, -66.5, -68.8,
        -77.3, -70.7, -76.5, -72.0,
        attackDuration = 0.050,
        decayDuration = 0.1,
        sustainLevel = 0.85,
        releaseDuration = 0.190,
        testReleaseTime = 0.7,    
    },
    
    {
        name = 'trumpet',
        -23.5, -15.3, -13.7, -19.5, -23.8, -25.2, -26.1, -35.9, -38.4, -48.4,
        -52.9, -56.6, -60.0, -61.5, -67.3, -71.6, -71.2,
        attackDuration = 0.050,
        decayDuration = 0.1,
        sustainLevel = 0.7,
        releaseDuration = 0.190,
        testReleaseTime = 0.7,    
    },
    
    {
        name = 'drum',
        -10.7, -27.6, -24.7,
        attackDuration = 0.0,
        decayDuration = 0.8,
        sustainLevel = 0.0,
        releaseDuration = 0.0,
        testReleaseTime = 0.5,          
    },
    
    {
        name = 'voice',
        -8.7, -16.0, -13.5, -35.0,
        attackDuration = 0.065,
        decayDuration = 0.2,
        sustainLevel = 0.6,
        releaseDuration = 0.15,
        testReleaseTime = 0.5,          
    },
    
    {
        name = 'accordion',
        -21.5, -12.2, -33.2, -34.0, -31.8, -19.0, -31.1, -30.7, -44.0, -33.6, 
        -47.7, -44.1, -47.8, -37.7, -46.6, -42.0, -74.3, -42.7, -52.5, -39.6,
        -54.9, -46.8, -61.5, -50.1, -59.6,
        attackDuration = 0.120,
        decayDuration = 0.8,
        sustainLevel = 0.5,
        releaseDuration = 0.120,
        testReleaseTime = 0.7,          
    },
    
    {
        name = 'xylophone',
        -11.6, -56.8, -64.5, -30.4, -66.3, -63.1, -52.8, -75.7, -74.8,
        attackDuration = 0.0,
        decayDuration = 0.350,
        sustainLevel = 0.0,
        releaseDuration = 0.0,
        testReleaseTime = 0.5,          
    },
    
    {
        name = 'soft',
        -9.1, -45.1, -32.3, -46.6, -53.3, -49.0, -64.8, -70.4,
        attackDuration = 0.400,
        decayDuration =  1.5,
        sustainLevel = 0.0,
        releaseDuration = 0.120,
        testReleaseTime = 0.8,
    },
}

---------------------------------------------------------------------------
-- wave generation (instrument synthesis)
---------------------------------------------------------------------------

-- generate wave as list of fp values
-- return list

function GenerateWave(instrument)
    local pi = math.pi
    local sin = math.sin

    assert(#instrument >= 1)
    
    -- convert harmonics table from dB to amplitude factors,
    -- using the fundamental frequency as reference (0 dB)
    -- (we are computing voltages, so 20 dB = a factor of 10)
    local hamplitude = {}
    local href = instrument[1]
    for n, hdb in ipairs(instrument) do
        hamplitude[n] = 10 ^ ((hdb - href) / 20)
    end
    
    -- angle increment for each step (for fundamental frequency)
    local astep = (2 * pi) / waveLength
    
    -- sum sine waves for fundamental and all harmonics
    local wave = {}
    for i = 1, waveLength  do
        local v = 0.0
        local idx = i - 1
        for n = 1, #instrument do
            -- harmonic n (f=1) by its amplitude
            v = v + (sin(idx * astep * n) * hamplitude[n])
        end
        wave[i] = v
    end
    
    return wave
end

---------------------------------------------------------------------------
-- normalize table values to given maximum value (maxval must be >=0)
-- return new normalized table

function NormalizeTable(t, maxval)
    local abs = math.abs

    -- get current max sample value (absolute value)
    local cmax = 0.0
    for _, v in ipairs(t) do
        local av = abs(v)
        if av > cmax then
            cmax = av
        end
    end

    -- normalize
    local tnorm = {}
    local scale = maxval / cmax
    for _, v in ipairs(t) do
        tnorm[#tnorm + 1] = v * scale
    end
    
    return tnorm
end

---------------------------------------------------------------------------
-- convert wave samples the +1/-1 range to +127/-127 range, 
-- round to nearest integer, convert to 2-complement signed integer
-- return converted wave as new table

function SamplesToBytes(wave)
    local floor = math.floor
    local byteWave = {}
    
    for _, v in ipairs(wave) do
        v = floor((v * 127) + 0.5)
        if v < 0 then
            -- 2-complement
            v = v + 256
        end
        -- just to be safe
        assert((v >= 0) and (v <= 255))
        byteWave[#byteWave + 1] = v
    end
    
    return byteWave
end

---------------------------------------------------------------------------
-- synthesize instrument wave
-- return wave as samples normalized to normLevel

function GenerateNormalizedWave(instrument, normLevel)
    -- generate base wave (single cycle) by sum of harmonics
    local baseWave = GenerateWave(instrument)
    -- normalize just below 1.0 to minimize clipping when converting to integers
    return NormalizeTable(baseWave, normLevel)
end

---------------------------------------------------------------------------
-- decay table generation
---------------------------------------------------------------------------

-- build exp. decay table from 1.0 (excluded) down to limit, 
-- set last entry to 0
-- (used for both decay and release)
-- return decay table

function BuildDecayTable(limit)
    assert((limit > 0) and (limit < 1.0))
    
    local decayTable = {}
    
    -- exp decay to last-but-one entry
    dt = adsrTableSize - 1
    lambda = -math.log(limit / 1.0) / dt
    for t = 1, dt do
        decayTable[#decayTable + 1] = math.exp(-lambda * t)
    end
    
    -- set last entry to 0
    decayTable[#decayTable + 1] = 0.0
    
    return decayTable
end

---------------------------------------------------------------------------
 -- test sound generation
---------------------------------------------------------------------------

-- build ADSR envelope as table with levels (0.0..1.0) for each adsrTimeStep
-- (release can happen even during decay)

function BuildAdsrEnvelope(instrument, decayTable)
    local envelope = {}
    
    -- shortcuts
    local floor = math.floor
    
    -- clamp ADSR values to max allowed
    local attackDuration = math.min(instrument.attackDuration, maxDuration)
    local decayDuration = math.min(instrument.decayDuration, maxDuration)
    local sustainLevel = math.min(instrument.sustainLevel, 1.0)
    local releaseDuration = math.min(instrument.releaseDuration, maxDuration)
    local testReleaseTime = math.min(instrument.testReleaseTime, maxDuration)
        
    -- attack (there must be no initial 0)
    if attackDuration > 0 then
        local numSteps = floor(attackDuration / adsrTimeStep)
        local levelStep = 1.0 / numSteps
        -- linear attack
        for n = 1, numSteps do
            envelope[#envelope + 1] = n * levelStep
        end
    end
    
    -- decay to sustain level, can be terminated by early release
    local elapsedTime = 0
    local levelDelta = 1.0 - sustainLevel
    -- index in decay table
    local decayIdx = 1
    while ((decayIdx <= #decayTable) and (elapsedTime < (testReleaseTime - attackDuration))) do
        envelope[#envelope + 1] = sustainLevel + (decayTable[decayIdx] * levelDelta)
        elapsedTime = elapsedTime + adsrTimeStep
        -- get index into exp decay table
        decayIdx = floor((elapsedTime / decayDuration) * #decayTable) + 1
    end

    -- if decay ended naturally (no early release), enter sustain mode
    if decayIdx > #decayTable then
        envelope[#envelope] = sustainLevel
        local totalTime = #envelope * adsrTimeStep
        while totalTime < testReleaseTime do
            -- (this will not happen on early release)
            envelope[#envelope + 1] = sustainLevel
            totalTime = totalTime + adsrTimeStep
        end
    end

    -- release from current level (final decay to zero)
    local initialLevel = envelope[#envelope]
    local elapsedTime = 0
    -- index in decay table
    local decayIdx = 1
    while decayIdx <= #decayTable do
        envelope[#envelope + 1] = decayTable[decayIdx] * initialLevel
        elapsedTime = elapsedTime + adsrTimeStep
        -- get index into exp decay table
        decayIdx = floor((elapsedTime / releaseDuration) * #decayTable) + 1
    end
    -- be sure last value is zero (may be missed in fast release)
    envelope[#envelope] = 0.0

    return envelope
end

---------------------------------------------------------------------------
-- generate envelope-modulated sound (for the duration of the envelope)
-- at the given multiple of the base frequency, assuming sampleFrequency,
-- add modulated sound to the given sound table

function AddModulateWave(sound, wave, envelope, freqMultiplier)
    local floor = math.floor

    -- duration of a wave cycle in seconds
    local waveDuration = #wave / (sampleFrequency * freqMultiplier)

    -- time 'played'
    local elapsedTime = 0.0
    -- fractonary index in wave table
    local waveIndex = 1.0
    -- fractionary step to advance wave index (for clarity)
    local waveIndexStep = freqMultiplier
    -- index in envelope table
    local envelopeIdx = 1
    
    -- reproduce the wave until the end of the envelope
    while envelopeIdx <= #envelope do
        -- get current level from envelope
        local level = envelope[envelopeIdx]

        -- insert a full wave cycle to mimic player behaviour
        while floor(waveIndex) <= #wave do
            sound[#sound + 1] = wave[floor(waveIndex)] * level
            waveIndex = waveIndex + waveIndexStep
        end
        -- wave index is beyond end, wrap it
        waveIndex = waveIndex - #wave
        
        -- update time and envelope index
        elapsedTime = elapsedTime + waveDuration
        envelopeIdx = floor(elapsedTime / adsrTimeStep) + 1
    end
end

---------------------------------------------------------------------------
-- add silence to sound table for the given duration (seconds) 
-- asssuming sampleFrequency

function AddSilence(sound, duration)
    local sampleCount = math.floor(sampleFrequency * duration)
    
    for i = 1, sampleCount do
        sound[#sound + 1] = 0
    end
end

---------------------------------------------------------------------------
-- generate test sound with the instrument sound (modulated by envelope)
-- at the given frequencies[]
-- return sound as new table

function GenerateTestSound(instrument, normalizedWave, decayTable, frequencies)
    local testSound = {}

    -- frequncy if played at 1:1 sample rate
    local baseFreq = sampleFrequency / waveLength
    
    -- use ADSR to build an envelope
    local envelope = BuildAdsrEnvelope(instrument, decayTable)

    -- modulate the normalized wave with the ADSR envelope
    -- at the different frequencies, with silence in between
    for n, freq in ipairs(frequencies) do
        AddModulateWave(testSound, normalizedWave, envelope, freq / baseFreq) 
        if n < #frequencies then
            AddSilence(testSound, testSilenceDuration)
        end
    end
    
    return testSound
end

---------------------------------------------------------------------------
-- output
---------------------------------------------------------------------------

-- write 8 bit values as C source lines

function WriteCBytes(bytes, indent, bytesPerLine)
    local sindent = string.rep(' ', indent)
    
    local byteCount = 0
    io.write(sindent)
    for n, v in ipairs(bytes) do
        if byteCount >= bytesPerLine then
            io.write('\n', sindent)
            byteCount = 0
        end
        io.write(string.format('0x%02x%s', v, (n < #bytes) and ', ' or ''))
        byteCount = byteCount + 1
    end
end

---------------------------------------------------------------------------
-- compute adsr table step from desired duration (s), limit to table size - 1
-- return step as 0xnnnn string representing (8 integer, 8 frac) hex value

function GetAdsrStep(duration)
    -- time to scan a 256-entry table at 1 entry/step
    local baseTableTime = adsrTimeStep * adsrTableSize
    
    -- ensure valid range (note: division may give #INF)
    duration = math.min(duration, maxDuration)
    local step = baseTableTime / duration
    if step >= adsrTableSize then
        step = adsrTableSize - 1
    end
    
    return string.format('0x%04x', math.floor(step * 256))
end

---------------------------------------------------------------------------
-- write decay table to stdout as 8-bit uint C array

function WriteDecayTable(decayTable)
    local floor = math.floor

    -- normalize to 0..255, just to be safe
    local dt = NormalizeTable(decayTable, 255)

    -- convert to nearest integer
    for n, v in ipairs(dt) do
        v = floor(v + 0.5)
        -- just to be safe
        assert((v >= 0) and (v <= 255))
        dt[n] = v
    end
    
    -- write as C table of unsigned bytes, 12 bytes per line
    io.write('const uint8_t adsrDecayTable[256] PROGMEM '
             .. '__attribute__ ((aligned (256))) = {\n')
    WriteCBytes(dt, 4, 12)
    io.write('\n};\n\n')
end

---------------------------------------------------------------------------
-- write instrument to stdout as C structure:
-- 8-bit wave array followed by ADSR parameters in player format

function WriteInstrumentStruct(waveBytes, instrument)
    io.write('const Instrument instrum_' .. instrument.name .. ' PROGMEM '
             .. '__attribute__ ((aligned (256))) = {\n')

    -- write wave bytes (signed), 10 bytes per line
    io.write('    .wave = {\n')
    WriteCBytes(waveBytes, 8, 10)
    io.write('\n    },\n')
    
    -- write ADSR parameters
    local sustainLevel = math.min(instrument.sustainLevel, 1.0)
    sustainLevel = math.floor((sustainLevel * adsrMaxLevel) + 0.5)
    io.write('    .attackStep = ', GetAdsrStep(instrument.attackDuration), ',\n')
    io.write('    .decayStep = ', GetAdsrStep(instrument.decayDuration), ',\n')
    io.write('    .sustainLevel = ', string.format('0x%02x', sustainLevel), ',\n')
    io.write('    .releaseStep = ', GetAdsrStep(instrument.releaseDuration), '\n')
    
    io.write('};\n\n')
end

---------------------------------------------------------------------------
-- write ordered instrument list to stdout as C array of pointers

function WriteInstrumentList()
    -- get longest name for alignment
    local maxlen = 0
    for _, instrum in ipairs(instruments) do
        maxlen = math.max(maxlen, #instrum.name)
    end
    
    io.write('InstrumentConstPtr instruments[NUM_INSTRUMENTS] PROGMEM = {\n')

    for n = 1, totalInstruments do
        -- use first instrument for undefined entries
        local instrum = instruments[n] or instruments[1]
        -- spacing required for alignment
    
        -- write pointer to intrument
        io.write('    &instrum_', instrum.name)
        io.write(( n < totalInstruments) and ',' or ' ')
        -- insert alignment spaces
        if #instrum.name < maxlen then
            io.write(string.rep(' ', maxlen - #instrum.name))
        end
        -- add 0-based instrument number as comment
        io.write(' // ', n - 1,'\n')
    end
    
    io.write('};\n')
end

---------------------------------------------------------------------------
-- write table samples to sampleFrequency, signed 8-bit .wav file
-- (using SoX instead of writing a .wav file writer)

function WriteWavFile(wave, filename)
    local byteToChar = string.char

    -- check if running under Windows
    -- (thanks to Oliver Kroth and others on the Lua mailing list)
    local isWindows = (string.sub(package.config, 1,1) == '\\')
    
    -- SoX command line for raw->wav conversion (8 bits, signed, 1 channel)
    local cmdline = 'sox -c 1 -r ' .. sampleFrequency 
                    .. ' -t raw -1 -s - -t wav ' .. filename
    
    -- launch SoX, connect to its standard input
    -- (note: arg and type of possible error depend on OS)
    local sox = io.popen(cmdline, isWindows and 'wb' or 'w')
    if sox then
        -- pipe opened ok, write wave samples
        for i, v in ipairs(wave) do
            sox:write(byteToChar(v))
        end

        -- close pipe
        sox:close()
    end
end

---------------------------------------------------------------------------
-- main program

-- (note: could move instruments to external file: pcall, check & clamp)

-- build table of instruments by name
local instrumentsByName = {}
for _, instrum in ipairs(instruments) do
    instrumentsByName[instrum.name] = instrum
end

-- get arg, must be either 'c', 'list' or a valid instrument name
local a = arg[1]
local printTables = a and (a:lower() == 'c')
local printList = a and (a:lower() == 'list')
local instrumentName = instrumentsByName[a] and a
if not (printTables or printList or instrumentName) then
    -- no name given, list available names
    io.stderr:write(version, '\n', usage)
    io.stderr:write('  valid names:\n')
    for i, instrum in pairs(instruments) do
        io.stderr:write('    (', i - 1, ') ', instrum.name, '\n')
    end
    os.exit(1)
end

-- check number of instruments
if #instruments > totalInstruments then
    io.stderr:write('- too many instruments (', #instruments,
        '/', totalInstruments, ')\n')
    os.exit(1)    
end

-- build table for exponential decay to a single level step
local decayTable = BuildDecayTable(1.0 / adsrMaxLevel)

if printTables then
    -- start generation of C source code (instrum.c)
    io.write('\n// --- program-generated file, do not edit ---\n')
    io.write('// --- see .h for license\n\n')
    io.write('#include "instrum.h"\n\n')
    -- write decay table as uint C table, 256-aligned, 12 bytes per line
    WriteDecayTable(decayTable)
    
    -- scan instruments
    for _, instrument in ipairs(instruments) do
        -- synthesize byte wave for the requested instrument
        -- (inefficient for historical reasons, could pre-compute harmonics)
        local normWave = GenerateNormalizedWave(instrument, normalizeLevel)
        local waveBytes = SamplesToBytes(normWave)
        -- write instrument data as C structure
        -- with ADSR values converted for player
        WriteInstrumentStruct(waveBytes, instrument)
    end
    
    -- write list of instruments as C array
    WriteInstrumentList()

elseif printList then
    for _, instrument in ipairs(instruments) do
        io.write(instrument.name, '\n')
    end

elseif instrumentName then
    io.write('generating test sound...')
    -- synthesize wave for the requested instrument
    local instrument = instrumentsByName[instrumentName]
    local normWave = GenerateNormalizedWave(instrument, normalizeLevel)
    -- generate test sound at multiple frequencies:
    local testSound = GenerateTestSound(instrument, normWave, decayTable, testFrequencies)
    -- convert to integer byte values in 2-complement representation
    local soundBytes = SamplesToBytes(testSound)
    -- write converted test sound to .wav file
    WriteWavFile(soundBytes, instrumentName .. '.wav')
    io.write('\n')
end

---------------------------------------------------------------------------
