---------------------------------------------------------------------------
-- music compiler for play-v6
-- by Enrico Colombini <erix@erix.it>
-- see musicomp.lua for license
--
-- module: action stream processor
-- see table at end of file for exported functions
---------------------------------------------------------------------------

-- access to public variables
local pairs = pairs
local ipairs = ipairs
local tonumber = tonumber
local tostring = tostring
local floor = math.floor
local assert = assert
local error = error
local table = table
local io = io

-- all the rest is private
local privateEnv = {}
_ENV = privateEnv

---------------------------------------------------------------------------
-- constants

-- note: these accents are perhaps a bit strong,
-- should be made configurable

-- main beat accents
beatAccents = {
    [2] = { 1.0, 0.6 },
    [3] = { 1.0, 0.6, 0.6 },
    [4] = { 1.0, 0.6, 0.8, 0.5 },
}

-- subdivision accents (compound is also used for triplets)
subdivAccents = {
    simple = { 1.0, 0.5 },
    compound = { 1.0, 0.5, 0.5 }, 
}

---------------------------------------------------------------------------
-- join actions from event list into a single action stream
-- sorted in timestamp order, preserve order within voices,
-- add 'timestamp' field to each action, set initial defaults
-- return the stream as list

function BuildActionStream(eventList)
    -- build a list of the timestamp keys used in the event list
    local timestamps = {}
    for k, v in pairs(eventList) do
        timestamps[#timestamps + 1] = k
    end
    -- sort list by increasing timestamp (note: voice, action orders are preserved)
    table.sort(timestamps)

    -- build action stream in timestamp order
    local actionStream = {}
    
    local prevTimestamp = 0
    for _, ts in ipairs(timestamps) do
        -- add wait action to get to timestamp ts, if needed
        if ts > prevTimestamp then
            actionStream[#actionStream + 1] = 
                { action = 'wait', duration = ts - prevTimestamp, timestamp = prevTimestamp}
        end
        prevTimestamp = ts

        -- add actions from the event at timestamp ts
        local event = eventList[ts]
        for _, action in ipairs(event) do
            -- write timestamp into action before adding to stream
            action.timestamp = ts
            actionStream[#actionStream + 1] = action
        end
    end
    
    -- timestamp- and voice-ordered list of action descriptions
    return actionStream
end

---------------------------------------------------------------------------
-- get duration of first measure (can be incomplete), 0 if no measures
-- return duration in 1/64

function GetFirstMeasureDuration(actionStream)
    local idx = 1
    local duration = 0
    while (idx < #actionStream) and (duration == 0) do
        local action = actionStream[idx]
        if action.action == 'end_measure' then
            duration = action.duration
        end
        idx = idx + 1
    end
    
    return duration
end

---------------------------------------------------------------------------
-- get time information from time signature string
-- (compound = true if compound time)
-- return time info in the format:
-- {
--     measureDuration = duration, -- in 1/64
--     numBeats = numBeats,
--     accents = { a1, a2, a3, ...} -- 0.0 .. 1.0, measureDuration entries
-- }

function GetTimeInfo(tsig, compound)
    -- get number of units and unit
    local numUnits, unit = tsig:match('(%d+)%/(%d+)')
    numUnits = tonumber(numUnits)
    
    -- convert unit to 1/64, get measure duration
    local unit64 = 64 / tonumber(unit)
    local measureDuration = unit64 * numUnits
    
    -- get number of beats per measure and beat subdivision duration
    local numBeats = compound and (numUnits / 3) or numUnits
    local beatSubdivDuration = compound and unit64 or (unit64 / 2)
    
    -- just to be safe, check that all values are integer
    assert(unit64 == floor(unit64))
    assert(numBeats == floor(numBeats))
    assert(beatSubdivDuration == floor(beatSubdivDuration))
    
    -- get appropriate accent tables
    local baseAccents = beatAccents[numBeats]
    local subAccents = subdivAccents[compound and 'compound' or 'simple']
    -- get number of beat subdivisions (2 or 3) for clarity
    local numBeatSubdiv = #subAccents

    -- build accents table for each 1/64 of measure
    local accents = {}
    for beat = 1, numBeats do
        -- base accent for this beat
        local baseBeatAccent = baseAccents[beat]
        -- consider beat subdivisions
        for subdiv = 1, numBeatSubdiv do
            -- base accent for this subdivision (combined with beat accent)
            local subdivAccent = baseBeatAccent * subAccents[subdiv]
            -- fill up subdivision accents
            for n = 1, beatSubdivDuration do
                accents[#accents + 1] = subdivAccent
            end
        end -- (subdiv)
    end -- (beat)
    
    -- just to be safe (#accents = numBeats * numBeatSubdiv * beatSubdivDuration)
    assert(#accents == measureDuration)
    
    return { measureDuration = measureDuration, 
             numBeats = numBeats, accents = accents }
end

---------------------------------------------------------------------------
-- preprocess combined action stream: add volume to note actions
-- considering main volume, mix, time signature accents, triplets

function PreprocessActionStream(actionStream, initialVolume)
    -- note: volume will be normalized to the highest used level
    local sectionVolume = initialVolume

    -- mix level for each voice, initially undefined
    local voiceMix = {}
    -- mix level default for each voice
    local voiceMixDefault = 1.0
    
    -- tempo, time signature undefined yet
    local tempo = nil
    local timeInfo = nil
    local accents = nil
    
    -- expected measure duration, use actual value for first measure
    -- (can be shorter than declared in time signature)
    local measureDuration = GetFirstMeasureDuration(actionStream)
    
    -- timestamp of start of current measure
    local measureStartTimestamp = nil

    -- accent at start of triplet (note or pause)
    local triplet1RefAccent = 0.0
    
    -- scan actions
    for _, action in ipairs(actionStream) do
        local act = action.action
        
        if act == 'main_volume' then
            -- set main volume, common to all voices
            sectionVolume = action.volume
        
        elseif act == 'voice_mix' then
            -- set mix level for this voice
            voiceMix[action.voice] = action.level
            
        elseif act == 'tempo' then
            -- tempo in quarter notes per minute
            tempo = action.qpm
            
        elseif act == 'time_signature' then
            -- get time signature info and accents
            timeInfo = GetTimeInfo(action.time, action.compound)
            accents = timeInfo.accents
            
            if not measureStartTimestamp then
                -- first measure: can get a negative starting timestamp
                measureStartTimestamp = measureDuration - timeInfo.measureDuration
            end
            -- from now on, this is the expected duration
            measureDuration = timeInfo.measureDuration
            
        elseif act == 'end_measure' then
            -- new measure starts
            measureStartTimestamp = action.timestamp
            
        elseif act == 'start_note' then
            if not tempo then
                error("tempo must be defined before the first note", 0)
            elseif not timeInfo then
                error("time signature must be defined before the first note", 0)
            end

            -- get accent for this timestamp
            local accent = accents[action.timestamp - measureStartTimestamp + 1] or 1.0
            
            -- triplet accent replaces 2nd, 3rd using stored1st as reference
            local triplet = action.triplet
            if (triplet == 2) or (triplet == 3) then
                accent = triplet1RefAccent * subdivAccents.compound[triplet]
            end

            -- set note volume considering main volume, mixer and accents
            -- use voice mixer default if not defined yet
            -- use defaults in case of accent table overflow (measure too long)
            local mix = voiceMix[action.voice] or voiceMixDefault
            action.volume = sectionVolume * mix * accent
        
        elseif act == 'triplet_1' then
            -- start of triplet: store reference accent
            triplet1RefAccent = accents[action.timestamp - measureStartTimestamp + 1] or 1.0
        
        end -- (act switch)

    end -- (scan actions)
end

---------------------------------------------------------------------------
-- normalize volumes for all notes in action stream to max = 1.0
-- (it would be more efficient to do this in PreprocessActionStream above,
-- but it is separated to avoid cluttering; performance cost is negligible)

function NormalizeStreamVolume(actionStream)
    -- highest volume level in stream
    local maxVolume = 0.0
    
    -- list of note actions in the stream
    local notes = {}

    -- scan stream for note actions
    for _, action in ipairs(actionStream) do
        if action.action == 'start_note' then
            -- add to list for later volume processing
            notes[#notes + 1] = action
            -- remember max volume
            if action.volume > maxVolume then
                maxVolume = action.volume
            end
        end
    end
    
    -- normalization factor to bring max volume to 1.0
    local normFactor = 1.0 / maxVolume
    
    -- normalize volume for all notes
    for _, action in ipairs(notes) do
        action.volume = action.volume * normFactor
    end
end

---------------------------------------------------------------------------
-- print contents of action stream (for testing)

function PrintActionStream(actionStream)
    for _, action in ipairs(actionStream) do
        -- show timestamp and action name first
        io.write(action.timestamp, ': ', action.action, ' ')
        
        -- show all other fields
        for k, v in pairs(action) do
            if (k ~= 'timestamp') and (k ~= 'action') then
                io.write(k, '=', tostring(v), ' ')
            end
        end
        
        io.write('\n')
    end
end

---------------------------------------------------------------------------
-- functions exported by this module 
---------------------------------------------------------------------------

return {
    BuildActionStream = BuildActionStream,
    PreprocessActionStream = PreprocessActionStream,
    NormalizeStreamVolume = NormalizeStreamVolume,
    PrintActionStream = PrintActionStream,
}

---------------------------------------------------------------------------
