---------------------------------------------------------------------------
-- music compiler for play-v6
-- by Enrico Colombini <erix@erix.it>
-- see musicomp.lua for license
--
-- module: music loader
-- see table at end of file for exported functions
---------------------------------------------------------------------------

-- access to public variables
local pairs = pairs
local ipairs = ipairs
local io = io
local error = error

-- all the rest is private
local privateEnv = {}
_ENV = privateEnv

---------------------------------------------------------------------------
-- source file format (lines starting with # are ignored)
-- note: a block can define a measure or just contain directives or commands

--   [sectionName]
--   block 1, one line per voice
--   empty line(s)
--   block 2, one line per voice
--   empty line(s)
--   ...
--   [other sectionName] (as above)

---------------------------------------------------------------------------
-- music table format:
--  {
--      -- list of sections in file order
--      sectionList = { section, section, section, ... };
--      -- sections, by name
--      sectionName = {
--          name = sectionName,
--          -- blocks, containing one line per voice
--          [1] = { linenum=num; [1]=block1voice1, [2]=block1voice2, ... },
--          [2] = { linenum=num; [1]=block2voice1, [2]=block2voice2, ... },
--          ...
--      },
--      -- (other sections, if any)
-- }

---------------------------------------------------------------------------
-- load music from source lines, organize into music table
-- return music table

function ReadMusicTable(musicLines)
    local music = { sectionList = {} }
    local currentSection = nil
    local currentBlock = nil
    
    for linenum, line in ipairs(musicLines) do
        if line:match('^%s*%[') then
            -- [name]: start of new section, terminate current block
            currentBlock = nil

            local sectionName = line:match('^%s*%[(%a[%w_]*)%]')
            if not sectionName then
                error('line ' .. linenum .. ': invalid section name', 0)
            elseif music[sectionName] then
                error('line ' .. linenum .. ': duplicated section name: ' .. sectionName, 0)
            else
                -- create new section, make it current
                currentSection = { name = sectionName}
                music[sectionName] = currentSection
                -- add to ordered list too, for human-friendly checking
                music.sectionList[#music.sectionList + 1] = currentSection
                -- not in a block
                currentBlock = nil
            end

        elseif line:match('^%s*$') then
            -- empty line: terminate block
            currentBlock = nil

        elseif line:match('^%s*%#') then
            -- comment line: ignore

        else
            -- voice line
            if not currentSection then
                error('line ' .. linenum .. ': missing section name', 0)
            else
                if not currentBlock then
                    -- first line of a block: add a new block (store linenum)
                    currentBlock = { linenum = linenum }
                    currentSection[#currentSection + 1] = currentBlock
                end
                -- add voice line to current block
                currentBlock[#currentBlock + 1] = line
            end
            
        end

    end -- for each line

    return music
end

---------------------------------------------------------------------------
-- load music from source file, main section is sectionName,
-- organize into music table, check music table,
-- return music table

function LoadMusic(musicFilename, maxVoices, sectionName)
    -- read file lines into table
    -- (no need to check for file existence, any error will be intercepted)
    local musicLines = {}
    for line in io.lines(musicFilename) do
        musicLines[#musicLines +1] = line
    end

    -- read music from table of lines
    local music = ReadMusicTable(musicLines)
    
    -- check for existing, non-empy requested section
    if not music[sectionName] then
        error('missing [' .. sectionName .. '] section', 0)
    elseif #music[sectionName] == 0 then
        error('[' .. sectionName .. '] section is empty', 0)
    end
    
    -- check max number of voices
    for _, section in ipairs(music.sectionList) do
        for _, block in ipairs(section) do
            if #block > maxVoices then
                error('line ' .. (block.linenum + #block - 1) 
                      .. ': too many voice lines (max allowed is ' 
                      .. maxVoices .. ')', 0)
            end
        end
    end
    
    return music
end

---------------------------------------------------------------------------
-- print music table (for testing)

function PrintMusicTable(music)
    for _, section in pairs(music.sectionList) do
        io.write('[', section.name, ']', '\n')
        for nb, block in ipairs(section) do
            io.write('block ', nb, ', line ', block.linenum, '\n')
            for nv, voiceline in ipairs(block) do
                io.write('  ', nv, ') ', voiceline, '\n')
            end
        end
        io.write('\n')
    end
end

---------------------------------------------------------------------------
-- functions exported by this module 
---------------------------------------------------------------------------

return {
    LoadMusic = LoadMusic,
    PrintMusicTable = PrintMusicTable,
}

---------------------------------------------------------------------------

