---------------------------------------------------------------------------
-- music compiler for play-v6
-- Enrico Colombini, created Jan 2014
-- license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
-- http://creativecommons.org/licenses/by-sa/4.0/
-- CRC-8 computation follows 'iButton' (Dallas/Maxim)
-- and is taken from avr-libc manual (see licenses.txt)
---------------------------------------------------------------------------

version = '- musicomp 1.0 - (C) 2014 Enrico Colombini'
usage = '- usage: musicomp <music_file> <output_file> [<section_name>]'

---------------------------------------------------------------------------
-- constants

-- maximum number of voices
maxVoices = 6

defaultSectionName = 'main'

---------------------------------------------------------------------------
-- check Lua version first

if not bit32 then
    io.write('- Lua 5.2 required\n')  -- note: add bit32 emulation for 5.3
    os.exit(1) --> exit
end

-- tell 'require' to look first into the directory containing this script
maindir = string.match(arg[0], '(.-)[^\\/]+$')
package.path = maindir .. '?.lua;' .. package.path

-- load required modules

loader = require('m_loader')
compiler = require('m_compiler')
streamer = require('m_astream')
coder = require('m_codegen')
writer = require('m_writer')

---------------------------------------------------------------------------
-- compile input filename, create output file (args already checked),
-- show compilation statistics

function CompileFile(musicFilename, outputFilename, sectionName)
    -- be sure not to overwrite source file
    if musicFilename == outputFilename then
        error('destination file cannot have the same name as input file', 0)
    end

    -- load music as line blocks into table, check for valid section
    local music = loader.LoadMusic(musicFilename, maxVoices, sectionName)
    --loader.PrintMusicTable(music) -- uncomment for testing
    
    -- the requested section is defined, compile it to event list
    local eventList = compiler.CompileSection(music[sectionName], maxVoices)
    
    -- convert timestamp-indexed event list to action stream
    local actionStream = streamer.BuildActionStream(eventList)
    --streamer.PrintActionStream(actionStream) -- uncomment for testing

    -- add end-of-music action
    local lastTimestamp = actionStream[#actionStream].timestamp
    actionStream[#actionStream + 1] = { action = 'end_music', timestamp = lastTimestamp }

    -- apply volume, accents and other effects
    streamer.PreprocessActionStream(actionStream, compiler.defaultVolume)
    -- normalize volume to use full dynamic range in player
    streamer.NormalizeStreamVolume(actionStream)
    --streamer.PrintActionStream(actionStream) -- uncomment for testing
   
    -- build final code from combined stream
    local sectionCode = coder.GenerateCode(actionStream)
    --coder.PrintCode(sectionCode) -- uncomment for testing
    
    -- write final output code
    writer.WriteOutputFile(outputFilename, sectionCode)
    
    -- show statistics
    io.write('music bytecode length: ', #sectionCode, '\n')
end

---------------------------------------------------------------------------
-- main
---------------------------------------------------------------------------

-- get and check args
local musicFilename = arg[1]
local outputFilename = arg[2]
local sectionName = arg[3] or defaultSectionName
if not (musicFilename and outputFilename) then
    io.write(version, '\n')
    io.write(usage, '\n')
    os.exit(1) --> exit
end

-- compile (catch errors)
local ok, r = pcall(
    function()
        -- read instrument codes from the directory of this file
        compiler.ReadInstrumentCodes(maindir)
        CompileFile(musicFilename, outputFilename, sectionName)
    end
) -- (pcall)

-- show error, if any
if not ok then
    io.stderr:write('- ', r or '?', '\n')
    os.exit(2) --> exit
end

-- all ok
os.exit(0)

---------------------------------------------------------------------------
