---------------------------------------------------------------------------
-- music packer for play-v6
-- created Jan 2014 by Enrico Colombini <erix@erix.it>
-- license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
-- http://creativecommons.org/licenses/by-sa/4.0/
---------------------------------------------------------------------------

version = '- musipack 1.0 - (C) 2014 Enrico Colombini'
usage = '- usage: musipack <code_list_file>'

---------------------------------------------------------------------------
-- scan list file, load each music codefile into a table,
-- return table (or just call error())
--
-- table format:
-- {
--     [1] = { name=filename, code=bytes },
--     [2] = { name=filename, code=bytes },
--     ...
-- }

function LoadCodeFiles(listFilename)
    local codetable = {}

    -- open list file
    local listfile = io.open(listFilename, 'r')
    if not listfile then
        error('cannot open list file: ' .. listFilename, 0)
    end
    
    -- load codefiles
    for fname in listfile:lines() do
        -- consider non-empty lines
        fname = fname:match('(%S*)%s*') -- remove terminator, if any (linux/win)
        if fname then        
            -- open music codefile
            local codefile = io.open(fname, 'rb')
            if not codefile then
                error('cannot open music code file: ' .. fname, 0)
            end
            
            -- read codefile content
            local code = codefile:read('*a')
            codefile:close()
            
            -- add entry
            codetable[#codetable + 1] = { name = fname, code = code }
        end
    end
    
    listfile:close()
    
    return codetable
end

---------------------------------------------------------------------------
-- write builtinMusic[] array
-- (all files, minus initial 'L' and final CRC-8, concatenated)
-- return table of starting offsets (0-based)

function WriteMusicArray(codetable, bytesPerLine)
    local indent = '    '
    
    -- offset (0-based index) of the starting byte of each code file
    local offsetTable = {}
    
    io.write('const uint8_t builtinMusic[] PROGMEM = {\n')
    
    local totalCount = 0
    local lineCount = 0
    io.write(indent)
    
    for n, codeinfo in ipairs(codetable) do
        offsetTable[n] = totalCount
    
        local bytes = codeinfo.code
        local isLastCode = (n == #codetable)
        
        -- skip first, last byte of each codefile
        for i = 2, (#bytes - 1) do
            if lineCount >= bytesPerLine then
                io.write('\n', indent)
                lineCount = 0
            end
            local isLastByte = isLastCode and (i == (#bytes - 1))
            local b = bytes:byte(i)
            io.write(string.format('0x%02x%s', b, isLastByte and '' or ', '))
            lineCount = lineCount + 1
            totalCount = totalCount + 1
        end
    end

    io.write('\n};\n')
    
    return offsetTable
end

---------------------------------------------------------------------------
-- write builtinMusicList[] array
-- with music names as comment (extension removed)

function WriteMusicList(codetable, offsetTable)
    local indent = '    '
    
    io.write('const uint16_t builtinMusicList[] PROGMEM = {\n')

    for i, codeinfo in ipairs(codetable) do
        -- get index of starting byte
        local idx = offsetTable[i]
    
        io.write(indent, '(const uint16_t)(builtinMusic + ',
                 string.format('0x%04x%s', idx, ')'),
                 (i < #codetable) and ',' or ' ')
        
        -- add file name as comment (remove extension, if any)
        local ext = string.match(codeinfo.name, '%.[^%.]*$') or ''
        local name = string.sub(codeinfo.name, 1, #codeinfo.name - #ext)
        io.write(' //', name, '\n')
    end

    io.write('};\n')
end

---------------------------------------------------------------------------
-- generate and print C source file including:
--   builtinMusic[] array
--   builtinMusicList[] array
--   builtinMusicCount constant

function WriteMusicFile(codetable)
    io.write('\n// --- program-generated file, do not edit ---\n')
    io.write('// --- see .h and music source files for licenses\n\n')
    io.write('#include "music.h"\n')
    io.write('#include <avr/pgmspace.h>\n\n')

    -- lumped music code data
    local offsetTable = WriteMusicArray(codetable, 12)
    io.write('\n')

    -- offsets to start of code files
    WriteMusicList(codetable, offsetTable)
    io.write('\n')

    -- number of music code files
    io.write('const uint8_t builtinMusicCount PROGMEM = ', #codetable, ';\n')
end

---------------------------------------------------------------------------
-- main
---------------------------------------------------------------------------

-- get and check args
local listFilename = arg[1]
if not (listFilename) then
    io.write(version, '\n')
    io.write(usage, '\n')
    os.exit(1) --> exit
end

-- load music code files (catch errors)
local ok, r = pcall(LoadCodeFiles, listFilename)

-- show error, if any
if not ok then
    io.stderr:write('- ', r or '?', '\n')
    os.exit(2) --> exit
end
-- else result is the loaded table
local codetable = r

-- generate and print C source file
WriteMusicFile(codetable)

-- all ok
os.exit(0)

---------------------------------------------------------------------------
