---------------------------------------------------------------------------
-- table generation for play-v6 (sampling increments, tempo)
-- created Dec 2013 by Enrico Colombini <erix@erix.it>
-- license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
-- http://creativecommons.org/licenses/by-sa/4.0/
---------------------------------------------------------------------------

version = '- tables 1.0 - (C) 2014 Enrico Colombini'

-- parameters (currently written in source code):

-- number of samples in a wave cycle
-- (very near to G3 sampled at 44100 Hz, convenient for testing)
waveSamplesPerCycle = 225

-- MCU clock frequency (Hz)
clockFreq = 16e6

-- play cycle count (clock cycles between samples)
playCycle = 512

---------------------------------------------------------------------------
-- constants

-- semitone multiplicative increment
semitoneMult = 2 ^ (1.0 / 12)
-- reference note (Hz)
A1Hz = 55.0
-- index of reference note in octave table (1-based)
A1pos = 10
-- note (semitone) names for octave 'n'
noteNames = {
    "Cn  / Don",
    "C#n / Do#n  / Dbn / Rebn",
    "Dn  / Ren",
    "D#n / Re#n  / Ebn / Mibn",
    "En  / Min",
    "Fn  / Fan",
    "F#n / Fa#n  / Gbn / Solbn",
    "Gn  / Soln",
    "G#n / Sol#n / Abn / Labn",
    "An  / Lan",
    "A#n / La#n  / Bbn / Sibn",
    "Bn  / Sin",
}

-- player rate in samples/s
playingRate = clockFreq / playCycle
-- multiplier to convert desided frequency to reading step
stepReadMult = waveSamplesPerCycle / playingRate

---------------------------------------------------------------------------
-- variables

-- frequencies for octave 1, from C1 (DO1) to B1 (SI1)
octave1Freqs = { }

---------------------------------------------------------------------------
-- compute 8.8 step to play the given frequency
-- return two values:
--   8.8 step as 16-bit unsigned (rounded to nearest),
--   rounding error in cents

function GetStepAndError(freq)
    -- compute step (approx. to nearest integer)
    local trueStep = stepReadMult * freq
    local step = math.floor(trueStep * 256 + 0.5)
    -- compute error in cents
    local freqErrorHz = ((step / 256 / stepReadMult) - freq)
    local centError = 100 * freqErrorHz / ((freq * semitoneMult) - freq)
    return step, centError
end

---------------------------------------------------------------------------
-- main program

-- compute frequencies for octave1 from semitone distances
for sd = -9, 2 do
    octave1Freqs[sd + 10] = A1Hz * semitoneMult ^ sd;
end

-- write #includes and parameters info

io.write('\n// --- program-generated file, do not edit ---\n')
io.write('// --- see .h for license\n\n')
io.write('\n')
io.write('#include "tables.h"\n\n')
io.write('/*\n')
io.write('    wave samples per cycle: ', waveSamplesPerCycle, '\n')
io.write('    CPU frequency: ', clockFreq, ' Hz\n')
io.write('    play cycle: ', playCycle, '\n')
io.write('    = play rate: ', clockFreq / playCycle, ' Hz\n')
io.write('*/\n\n')

-- start of C note table
io.write('const uint16_t notes[64] PROGMEM = {\n')

-- generate note table for 64 notes (>5 octaves) from C1 up
-- (this also covers bass guitar with lowered tuning)

local octaveIdx = 1
local octave = 1
local octaveMult = 1
for noteIdx = 0, 63 do
    local f = octave1Freqs[octaveIdx] * octaveMult
    local step, errCents = GetStepAndError(f)
    
    -- generate entry line
    io.write(string.format('    0x%04x%s // [%2d] %8.3f (%+4.1f cents)  ', 
             step, ((noteIdx == 63) and ' ' or ','), noteIdx, f, errCents))
    io.write((noteNames[octaveIdx]:gsub('n', tostring(octave)))) -- only 1st result
    io.write('\n')
    
    -- next note (semitone), wrap around at octave change
    octaveIdx = octaveIdx + 1
    if octaveIdx > 12 then
        octaveIdx = 1
        octave = octave + 1
        octaveMult = octaveMult * 2
    end
end

-- end of C note table
io.write('};\n\n')

-- start of C tempo table
io.write('const uint16_t tempo64Steps[256] PROGMEM = {\n')

-- generate 16-bit tempo64 steps for 1/4 tempo = 1..256
-- a tempo64 step is the number of play cycles in 1/64, that is in 1/16 of 1/4
-- values > 0xffff are set to 0xff (for very low tempo)

for tempo = 1, 256 do
    -- compute step (approx. to nearest integer)
    local step64 = (playingRate * 60) / (tempo * 16)
    step64 = math.floor(step64 + 0.5)
    
    -- keep inside 16 bit
    if step64 > 0xffff then
        step64 = 0xffff
    end
    
    -- generate entry line
    io.write(string.format('    0x%04x%s // [%2d]\n', 
             step64, ((tempo == 256) and ' ' or ','), tempo))
end

-- end of C tempo table
io.write('};\n')

---------------------------------------------------------------------------
